import ee
import numpy as np
from loguru import logger
from sklearn.neural_network import MLPRegressor
from sklearn.preprocessing import StandardScaler


class eeMLPRegressor:
    """A custom class to exectue the .predict of scikit-learn's MLPRegressor on Google Earth Engine (GEE) server side.
    Requires a pretrained MLPRegressor an an input, extracts the weights and biases, import them into GEE runs the forward method on server side.


    Parameters
    ----------
    model : MLPRegressor
        An instance of the scikit-learn MLPRegressor class that has been fitted to data.

    trait_name : str, optional
        Name of the predicted trait. Default is 'trait'.

    Attributes
    ----------
    model : MLPRegressor
        The provided MLPRegressor instance.

    trait_name : str
        Name of the predicted trait.

    ee_array_weights : List[ee.Array]
        Weights of the neural network layers, converted to GEE ee.Array objects.

    ee_array_biases : List[ee.Array]
        Biases of the neural network layers, converted to GEE ee.Array objects.

    ee_image_weights : List[ee.Image]
        Weights of the neural network layers, converted to GEE ee.Image objects.

    ee_image_biases : List[ee.Image]
        Biases of the neural network layers, converted to GEE ee.Image objects.

    Methods
    -------
    _apply_activation_function(x: ee.Image or ee.Array, function: str) -> ee.Image or ee.Array:
        Apply the specified activation function to the input.

    _forward_pass_array(ee_X: ee.Array) -> ee.Array:
        Perform a forward pass on an input array through the neural network layers.

    _forward_pass_image(ee_X: ee.Image) -> ee.Image:
        Perform a forward pass on an input image through the neural network layers.

    predict(image: ee.Image or ee.Array) -> ee.Image:
        Predict an input image or array using the trained neural network.

    Notes
    -----
    TODO's:
    - Only regression implemented so far. Think about creating a similar eeMLPClassifier class. Both classes could inherit some their function from the base class
    - Check if activation functions are correctly implemented
    - Write test cases

    """

    def __init__(self, model: MLPRegressor, trait_name: str = "trait") -> None:
        self.model = model
        self.trait_name = trait_name
        # if predicting on an array
        self.ee_array_weights = [
            ee.Array(weights.tolist()) for weights in self.model.coefs_
        ]
        self.ee_array_biases = [
            ee.Array(biases.tolist()).reshape([1, -1])
            for biases in self.model.intercepts_
        ]
        # if predictin on an image
        self.ee_image_weights = [ee.Image(array) for array in self.ee_array_weights]
        self.ee_image_biases = [ee.Image(array) for array in self.ee_array_biases]
        # activation function
        # if(self.model.activation != 'tanh'):
        #     raise NotImplementedError
        if self.model.out_activation_ != "identity":
            raise NotImplementedError

    def _apply_activation_function(self, x: ee.Image, function: str) -> ee.Image:
        if function == "tanh":
            return x.tanh()
        if function == "identity":
            return x
        if function == "softmax":
            raise NotImplementedError
        if function == "logistic" or function == "sigmoid":
            return x.multiply(-1).exp().add(1).pow(-1)  # 1 / (1 + exp(-x))
        if function.lower() == "relu":
            return x.gt(0).multiply(x)  # max(0, x)
        raise ValueError

    def _forward_pass_array(self, ee_X):
        # this method is supposed to work with a array where the rows correspond to the number of rows, while the column cooresponds to the number of bands
        n_samples, _ = np.array(ee_X.getInfo()).shape
        x = ee_X  # dim: (n_samples, b_bands)
        for i in range(self.model.n_layers_ - 1):
            x = x.matrixMultiply(self.ee_array_weights[i])
            x = x.add(self.ee_array_biases[i].repeat(axis=0, copies=n_samples))
            if i != self.model.n_layers_ - 2:
                x = self._apply_activation_function(x, self.model.activation)
        # apply output activation
        x = self._apply_activation_function(x, self.model.out_activation_)
        return x

    def _forward_pass_image(self, ee_X: ee.Image) -> ee.Image:
        # # convert input image to arrayImage where each pixel holds an array of shape (1, n_bands)
        # array_X_1D = ee_X.toArray() # dim: (n_bands)
        # array_X_2D = array_X_1D.toArray(1) # dim: (n_bands, 1)
        # array_X_2D = array_X_2D.arrayTranspose() # dim: (1, n_bands)
        x = ee_X.toArray().toArray(1).arrayTranspose()  # dim: (1, n_bands)

        for i in range(self.model.n_layers_ - 1):
            x = x.matrixMultiply(
                self.ee_image_weights[i],
            )  # 1st iteration: dim (1, n_bands) x (n_bands, n_nodes)
            x = x.add(self.ee_image_biases[i])  # (1 x n_nodes)

            if i != self.model.n_layers_ - 2:
                x = self._apply_activation_function(x, self.model.activation)

        # apply output activation
        x = self._apply_activation_function(x, self.model.out_activation_)
        # convert back to image with single band with output trait value
        output_image = x.arrayProject([0]).arrayFlatten([[self.trait_name]])
        return output_image

    def predict(
        self,
        image: ee.Image,
        copy_properties: list[str] | None = None,
    ) -> ee.Image:
        if isinstance(image, ee.Image):
            # that should work now
            if copy_properties is not None:
                return self._forward_pass_image(image).copyProperties(
                    source=image,
                    properties=copy_properties,
                )
            return self._forward_pass_image(image)
        # elif isinstance(image, ee.Array):
        #     raise NotImplementedError
        raise TypeError


class eeStandardScaler:
    def __init__(self, scaler: StandardScaler, feature_names: list[str] | None = None):
        self.scaler_ = scaler
        self.mean_ = scaler.mean_
        self.scale_ = scaler.scale_
        self.ee_mean_ = ee.Array(self.mean_.tolist())  # dim: should be 1D: (n_bands)
        self.ee_scale_ = ee.Array(self.scale_.tolist())  # dim: should be 1D: (n_bands)
        self.feature_names_ = feature_names
        self.n_features_in_ = scaler.n_features_in_
        self.n_samples_seen_ = scaler.n_samples_seen_

        if self.feature_names_ is None:
            logger.trace(
                "No feature_names provided; they will be set to the names scaler.feature_names_in_",
            )
            self.feature_names_ = list(scaler.feature_names_in_)
        elif not len(self.feature_names_) == len(self.mean_):
            logger.error(
                f"Length of feature_names: {len(self.feature_names_)} does not match length of mean: {len(self.mean_)}",
            )
            raise ValueError

        # Create ee.Images for the mean and scale
        self.scaler_mean_image = ee.Image.constant(list(self.mean_)).rename(
            self.feature_names_,
        )
        self.scaler_scale_image = ee.Image.constant(list(self.scale_)).rename(
            self.feature_names_,
        )

    def transform_image(self, image: ee.Image) -> ee.Image:
        # Select the relevant bands
        image_selected = image.select(self.feature_names_)

        # Subtract the mean image and divide by the scale image
        image_scaled = image_selected.subtract(self.scaler_mean_image).divide(
            self.scaler_scale_image,
        )

        # Rename the scaled bands to match the original band names
        image_scaled = image_scaled.rename(self.feature_names_)

        # Add the scaled bands to the original image, replacing the original bands
        image_to_return = image.addBands(image_scaled, overwrite=True)

        return image_to_return

    def inverse_transform_column(self, image: ee.Image, column: str) -> ee.Image:
        if column not in self.feature_names_:
            raise ValueError(
                f"Column {column} not in feature names: {self.feature_names_}",
            )

        # get index of column in feature_names
        column_index = self.feature_names_.index(column)

        image_backtransformed = image.select(column).multiply(
            self.ee_scale_.get([column_index]),
        )
        image_backtransformed = image_backtransformed.add(
            self.ee_mean_.get([column_index]),
        )

        image_to_return = image.addBands(image_backtransformed, overwrite=True)
        return image_to_return


class eeMinMaxRangeMasker:
    def __init__(self, min_max_dict: dict, tolerance: float = 0.01):
        """A class to handle the detection of out-of-range values for satellite image bands
        using Google Earth Engine. It sets up tolerance levels for minimum and maximum values
        from a dictionary of predefined minimum and maximum values for each band and masks out-of-range
        pixels in an image.

        Attributes:
            min_max_dict (dict): A dictionary containing the minimum and maximum values for each band.
            tolerance (float): A percentage tolerance level for the minimum and maximum values.

        Methods:
            ee_image_min_max_masking(image: ee.Image) -> ee.Image:
                Masks out-of-range pixels based on predefined tolerance ranges for each band.

        """
        self.min_max_dict = min_max_dict
        self.min_vals = {k: v["min"] for k, v in min_max_dict.items()}
        self.max_vals = {k: v["max"] for k, v in min_max_dict.items()}
        self.ranges = {k: v["max"] - v["min"] for k, v in min_max_dict.items()}

        # Tolerance thresholds
        self.min_vals_tolerance = {
            k: v - self.ranges[k] * tolerance for k, v in self.min_vals.items()
        }
        self.max_vals_tolerance = {
            k: v + self.ranges[k] * tolerance for k, v in self.max_vals.items()
        }

        # Convert to Earth Engine objects
        self.ee_min_vals = ee.Dictionary(self.min_vals)
        self.ee_max_vals = ee.Dictionary(self.max_vals)
        self.ee_min_vals_tolerance = ee.Dictionary(self.min_vals_tolerance)
        self.ee_max_vals_tolerance = ee.Dictionary(self.max_vals_tolerance)

        self.ee_min_tolerance_image = ee.Image.constant(
            list(self.min_vals_tolerance.values()),
        ).rename(list(self.min_vals_tolerance.keys()))

        self.ee_max_tolerance_image = ee.Image.constant(
            list(self.max_vals_tolerance.values()),
        ).rename(list(self.max_vals_tolerance.keys()))

        self.band_names = list(self.min_max_dict.keys())
        self.ee_columns = ee.List(self.band_names)

    def ee_mask(self, image):
        selected_image = image.select(self.band_names)
        masked_image = selected_image.updateMask(
            selected_image.gte(self.ee_min_tolerance_image).And(
                selected_image.lte(self.ee_max_tolerance_image),
            ),
        )
        return image.addBands(masked_image, overwrite=True)



# run_export.py
from __future__ import annotations

from pathlib import Path

import typer
import yaml
from loguru import logger

from gee_biophys.config import ConfigParams
from gee_biophys.s2_export import export_image
from gee_biophys.s2_input import load_s2_input
from gee_biophys.s2_predict import biophys_predict
from gee_biophys.utils import (
    get_system_index,
    initialize_export_location,
    update_image_metadata,
)


def validate_config(path: Path):
    if not path or not path.exists():
        raise typer.BadParameter(
            "Please provide a valid path to a YAML configuration file, e.g.:\n"
            "  gee-biophys --config example_configs/minimal_example.yaml",
        )
    return path


def load_params(path: str) -> ConfigParams:
    with open(path) as f:
        data = yaml.safe_load(f)
    return ConfigParams(**data)


app = typer.Typer(
    no_args_is_help=True,
    help="Custom time-series export of LAIe/FAPAR/FCOVER with YAML params.",
)


def run_pipeline(config: str, set_public: bool = False) -> None:
    """Run the full export pipeline based on the provided configuration."""
    # <---- Setup ---->
    cfg = load_params(str(config))

    initialize_export_location(cfg, set_public=set_public)

    # <---- Main Loop ---->
    for interval_start, interval_end in cfg.temporal.iter_date_ranges():
        # <---- Load Input ---->
        logger.info(
            f"Processing interval: {interval_start.strftime('%Y-%m-%d')} to {interval_end.strftime('%Y-%m-%d')}",
        )
        imgc = load_s2_input(cfg, interval_start, interval_end)

        # <---- Prediction ---->
        output_image = biophys_predict(cfg, imgc)

        # <---- Update metadata ---->
        output_image = update_image_metadata(
            output_image,
            interval_start,
            interval_end,
            cfg,
        )
        filename = get_system_index(cfg, interval_start, interval_end)

        # <---- Export  ---->
        export_image(output_image, filename, cfg)

    logger.info(
        "Done! Exports have been started. Please check task status to see if they completed successfully.",
    )


@app.command(help="Run the export using the provided YAML configuration.")
def run(
    config: Path = typer.Option(
        ...,
        "--config",
        "-c",
        help="Path to params YAML file.",
        callback=validate_config,
        readable=True,
    ),
    public: bool = typer.Option(
        False,
        "--public/--no-public",
        help="Set exported file ACL to public after export. This is required if you want to visualize the exported asset using the GEE-App ",
        show_default=True,
    ),
):
    """CLI entry point"""
    typer.echo(f"Using configuration file: {config}")
    run_pipeline(config, set_public=public)


if __name__ == "__main__":
    config_path = Path("example_configs/forest-fire-bitsch-2023.yaml")
    run_pipeline(config_path, set_public=True)  # for debugging purposes



from functools import reduce

import ee

from gee_biophys.config import ConfigParams
from gee_biophys.models.s2biophys import eePipelinePredictMap, load_model_ensemble
from gee_biophys.models.sl2p import load_SL2P_model
from gee_biophys.utils_predict import (
    aggregate_ensemble_predictions,
    aggregate_imagecollection_simple,
)


def biophys_predict(cfg: ConfigParams, input_imgc: ee.ImageCollection) -> ee.Image:
    """Apply the selected biophysical model to the input Sentinel-2 ImageCollection
    and return an ImageCollection with predicted biophysical variables.
    """
    if cfg.variables.model == "sl2p":
        model_mean, model_std = load_SL2P_model(variable=cfg.variables.variable)

        pred_mean_imgc = input_imgc.map(lambda img: model_mean.ee_predict(img))
        pred_std_imgc = input_imgc.map(lambda img: model_std.ee_predict(img))

        output_image = aggregate_ensemble_predictions(
            pred_mean_imgc,
            pred_std_imgc,
            cfg.variables.variable,
        )

    elif cfg.variables.model == "s2biophys":
        s2biophys_model_ensemble = load_model_ensemble(cfg.variables.variable)

        gee_preds = {}
        for i, (model_name, model) in enumerate(s2biophys_model_ensemble.items()):
            gee_preds[model_name] = eePipelinePredictMap(
                pipeline=model.pipeline,
                imgc=input_imgc,
                trait=cfg.variables.variable,
                model_config=model.config,
                min_max_bands=model.min_max_bands,
                clip_min_max=cfg.options.clip_min_max,
            )

        s2biophys_imgc_preds = reduce(
            lambda x, y: x.merge(y),
            gee_preds.values(),
        )

        output_image = aggregate_imagecollection_simple(
            s2biophys_imgc_preds,
            cfg.variables.variable,
            replications=len(s2biophys_model_ensemble),
        )

    else:
        raise ValueError(f"Unsupported model: {cfg.variables.model}")

    # select only desired output bands
    output_band_names = [
        f"{cfg.variables.variable}_{band}" for band in cfg.variables.bands
    ]

    return output_image.select(output_band_names)



import ee


def aggregate_imagecollection_simple(
    imgc: ee.ImageCollection,
    trait_name: str,
    replications: int = 1,
) -> ee.Image:
    """Aggregate an ImageCollection by calculating the mean, standard deviation, and count of images.

    Args:
        imgc (ee.ImageCollection): The input ImageCollection containing predictions.
        trait_name (str): The name of the trait being predicted (used for naming output bands).
        replications (int, optional): The number of model replications used to generate the ImageCollection. Defaults to 1.

    Returns:
        ee.Image: An Image containing the mean, standard deviation, and count of predictions.

    Notes:
       - The count band is normalized by the number of replications to reflect the proportion of valid predictions.

    """
    mean_name = f"{trait_name}_mean"
    std_name = f"{trait_name}_stdDev"
    count_name = f"{trait_name}_count"

    img_mean = imgc.mean().rename(mean_name)
    img_std = imgc.reduce(ee.Reducer.stdDev()).rename(std_name)
    img_counts = (
        imgc.reduce(ee.Reducer.count()).divide(replications).toInt().rename(count_name)
    )

    # TODO: min max range?
    img_to_return = ee.Image([img_mean, img_std, img_counts])
    return img_to_return


def aggregate_ensemble_predictions(
    mean_imgc: ee.ImageCollection,
    std_imgc: ee.ImageCollection,
    trait_name: str,
) -> ee.Image:
    mean_name = f"{trait_name}_mean"
    std_name = f"{trait_name}_stdDev"
    count_name = f"{trait_name}_count"

    img_mean = mean_imgc.mean().rename(mean_name)

    img_within_var = std_imgc.map(lambda img: img.pow(2)).mean()
    img_between_var = mean_imgc.reduce(ee.Reducer.variance())

    img_total_var = img_within_var.add(img_between_var)
    img_total_std = img_total_var.sqrt().rename(std_name)

    img_counts = mean_imgc.count().rename(count_name)
    img_to_return = ee.Image([img_mean, img_total_std, img_counts])
    return img_to_return



import numpy as np
import pandas as pd
from loguru import logger
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import ColumnTransformer, TransformedTargetRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler


# Custom transformer that converts specified columns (angles) to their cosines
class AngleTransformer(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        # No fitting necessary for cosine transformation
        return self

    def transform(self, X):
        # X is a numpy array here, not a DataFrame
        return np.cos(np.deg2rad(X))


class NIRvTransformer(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        # No fitting necessary for cosine transformation
        return self

    def transform(self, X):
        assert X.shape[1] == 10, f"Expected 10 bands, got {X.shape[1]}"
        # problem:
        # X is a numpy array here, not a DataFrame

        # add random number to avoid division by zero
        X = X + np.abs(np.random.rand(*X.shape) * 1e-10)

        if isinstance(X, np.ndarray):
            idx_nir = 6  # B8
            idx_red = 2  # B4
            NIR = X[:, idx_nir]
            RED = X[:, idx_red]
        elif isinstance(X, pd.DataFrame):
            NIR = X[["B8"]].values.reshape(-1)
            RED = X[["B4"]].values.reshape(-1)

        # check that NIR + RED is not 0
        if np.any(NIR + RED == 0):
            logger.warning("NIR + RED is 0 for some samples, set to mean sum")
            NIR[NIR + RED == 0] = np.mean(NIR)
            RED[NIR + RED == 0] = np.mean(RED)

        NDVI = (NIR - RED) / (NIR + RED)
        NIRv = NDVI * NIR

        if np.any(NIRv == 0):
            logger.warning(
                "NIRv is 0 for some samples, setting to mean of non-zero values",
            )
            NIRv[NIRv == 0] = np.mean(NIRv[NIRv != 0])

        # divide each row by the corresponding NIRv value
        X_normalized = X / NIRv[:, np.newaxis]
        return X_normalized


# Define the safe logit function
def safe_logit(x):
    # add warning when this clipping was used
    if np.any(x <= 1e-9) or np.any(x >= 1 - 1e-9):
        logger.trace("Clipping values to avoid 0 and 1")
        x = np.clip(x, 1e-9, 1 - 1e-9)  # Clip the values to avoid 0 and 1
    return np.log(x / (1 - x))


# Define the inverse sigmoid function
def safe_inverse_logit(x):
    return 1 / (1 + np.exp(-x))


def identity(x):
    return x


def get_pipeline(model: BaseEstimator, config: dict) -> Pipeline:
    angles = ["tts", "tto", "psi"]
    bands = ["B2", "B3", "B4", "B5", "B6", "B7", "B8", "B8A", "B11", "B12"]

    angle_transformer = Pipeline(steps=[("angle_transformer", AngleTransformer())])
    if config["nirv_norm"]:
        band_transformer = Pipeline(
            steps=[
                ("nirv_transformer", NIRvTransformer()),
                ("scaler", StandardScaler()),
            ],
        )
    else:
        band_transformer = Pipeline(steps=[("scaler", StandardScaler())])

    preprocessor = ColumnTransformer(
        transformers=[
            ("band_transformer", band_transformer, bands),
            ("angle_transformer", angle_transformer, angles),
        ],
        remainder="passthrough",
    )

    if config["transform_target"] == "log1p":
        regressor = TransformedTargetRegressor(
            regressor=model,
            func=np.log1p,
            inverse_func=np.expm1,
        )
    elif config["transform_target"] == "standard":
        regressor = TransformedTargetRegressor(
            regressor=model,
            transformer=StandardScaler(),
        )
    elif config["transform_target"] == "None":
        regressor = TransformedTargetRegressor(
            regressor=model,
            func=identity,
            inverse_func=identity,
        )
    elif config["transform_target"] == "logit":
        regressor = TransformedTargetRegressor(
            regressor=model,
            func=safe_logit,
            inverse_func=safe_inverse_logit,
        )
    else:
        raise ValueError(f"Unknown target transformation: {config['transform_target']}")

    pipeline = Pipeline(
        steps=[
            ("preprocessor", preprocessor),
            ("regressor", regressor),
        ],
    )

    return pipeline



import ee
import numpy as np
import pandas as pd
import pytest

from gee_biophys.models.s2biophys import eePipelinePredictMap, load_model_ensemble

ATOL = 1e-5
N_SAMPLES = 10


@pytest.mark.parametrize("variable", ["laie", "fapar", "fcover"])
def test_s2biophys_random_inputs(ee_init, variable):
    s2biophys_ensemble = load_model_ensemble(variable)

    rng = np.random.default_rng(123)
    band_order = [
        "B2",
        "B3",
        "B4",
        "B5",
        "B6",
        "B7",
        "B8",
        "B8A",
        "B11",
        "B12",
        "tts",
        "tto",
        "psi",
    ]

    # Sort models by name for deterministic behavior
    models = sorted(s2biophys_ensemble.items(), key=lambda kv: kv[0])  # (name, spec)

    for _ in range(N_SAMPLES):
        # -------- input sample (kept within reasonable valid ranges) --------
        sample = {}
        for b in band_order:
            if b.startswith("B"):
                sample[b] = float(rng.uniform(0.01, 0.4))
            elif b in ("tts", "tto"):
                sample[b] = float(rng.uniform(0, 60))
            elif b == "psi":
                sample[b] = float(rng.uniform(0, 180))

        df = pd.DataFrame([[sample[b] for b in band_order]], columns=band_order)

        # -------- local (sklearn) predictions, per model --------
        preds_np = np.array(
            [spec["pipeline"].predict(df).item() for _, spec in models], dtype=float
        )
        mean_np = float(preds_np.mean())
        std_np = float(preds_np.std())

        # -------- server (EE) predictions, per model, single round-trip --------
        # Build a constant image with our sample values, in correct band order
        img_in = ee.Image.constant([sample[b] for b in band_order]).rename(band_order)

        # Some pipelines expect an ImageCollection; wrap and pass through each model
        imgc_in = ee.ImageCollection([img_in])

        # Run each model; turn each result IC into a single-band Image named p_<i>
        pred_imgs = []
        for i, (_, spec) in enumerate(models):
            ic_pred = eePipelinePredictMap(
                pipeline=spec["pipeline"],
                imgc=imgc_in,
                trait=variable,
                model_config=spec["config"],
                min_max_bands=None,
                clip_min_max=False,
            )
            # Use mosaic() in case the pipeline returns more than one tile
            pred_imgs.append(
                ee.Image(ic_pred.mosaic()).select(variable).rename(f"p_{i}")
            )

        # Stack all per-model predictions into one multi-band image
        stack = ee.Image.cat(pred_imgs)

        # Compute ensemble mean/std on the stack server-side (across bands)
        mean_img = stack.reduce(ee.Reducer.mean()).rename(f"{variable}_mean")
        std_img = stack.reduce(ee.Reducer.stdDev()).rename(f"{variable}_stdDev")
        out_img = ee.Image.cat([stack, mean_img, std_img])

        # Sample once (use a tiny buffer+dropNulls to dodge mask edge-cases)
        pt = ee.Geometry.Point([0, 0])
        fc = out_img.sample(
            region=pt.buffer(100), scale=20, numPixels=1, dropNulls=True, tileScale=2
        )
        size = fc.size().getInfo()

        if size == 0:
            # As a fallback (should rarely trigger with constant input), use reduceRegion
            vals = out_img.reduceRegion(
                reducer=ee.Reducer.first(),
                geometry=pt,
                scale=20,
                bestEffort=True,
                maxPixels=1e13,
            ).getInfo()
        else:
            vals = ee.Feature(fc.first()).toDictionary().getInfo()

        # Pull per-model EE preds in consistent order
        preds_ee = np.array([vals[f"p_{i}"] for i in range(len(models))], dtype=float)
        mean_ee = float(vals[f"{variable}_mean"])
        std_ee = float(vals[f"{variable}_stdDev"])

        # -------- assertions with helpful diagnostics --------
        try:
            np.testing.assert_allclose(preds_np, preds_ee, atol=ATOL)
        except AssertionError as e:
            diffs = (preds_np - preds_ee).tolist()
            msg = (
                f"\nPer-model mismatch:\n"
                f"  local: {preds_np.tolist()}\n"
                f"   ee  : {preds_ee.tolist()}\n"
                f"  diff : {diffs}\n"
            )
            raise AssertionError(msg) from e

        try:
            np.testing.assert_allclose(mean_np, mean_ee, atol=ATOL)
            np.testing.assert_allclose(std_np, std_ee, atol=ATOL)
        except AssertionError as e:
            msg = (
                f"\nEnsemble mismatch:\n"
                f"  mean  local={mean_np:.8f}, ee={mean_ee:.8f}, Δ={mean_np - mean_ee:.2e}\n"
                f"  std   local={std_np:.8f},  ee={std_ee:.8f},  Δ={std_np - std_ee:.2e}\n"
                f"  per-model local: {preds_np.tolist()}\n"
                f"  per-model ee   : {preds_ee.tolist()}\n"
            )
            raise AssertionError(msg) from e



from datetime import datetime, timedelta

import ee
from dateutil.relativedelta import relativedelta
from loguru import logger

from gee_biophys.config import ConfigParams


def get_system_index(
    cfg: ConfigParams,
    interval_start: datetime,
    interval_end: datetime,
) -> str:
    band_suffix = "-".join(cfg.variables.bands)
    output_resolution = cfg.export.scale
    start_str = interval_start.strftime("%Y%m%d")
    end_str = interval_end.strftime("%Y%m%d")
    crs_str = cfg.export.crs.lower().replace("epsg:", "epsg-")
    version_str = cfg.version
    region_str = (
        cfg.spatial.region_name if cfg.spatial.region_name else cfg.spatial.type
    )

    index_str = (
        f"{cfg.variables.variable}_"
        f"{cfg.variables.model}_"
        f"{band_suffix}_"
        f"{output_resolution}m_s_"
        f"{start_str}_"
        f"{end_str}_"
        f"{region_str}_"
        f"{crs_str}_"
        f"{version_str}"
    )

    assert len(index_str) <= 100, (
        f"Generated system:index is too long ({len(index_str)} characters). {index_str} "
        "Please use shorter names for region_name in the configuration."
    )
    return index_str


def _get_cfg_string(cfg: ConfigParams) -> str:
    """Generate a string representation of the configuration parameters.
    This can be used for metadata or logging purposes.
    """
    cfg_parts = [
        f"spatial: {cfg.spatial}",
        f"temporal: {cfg.temporal}",
        f"variables: {cfg.variables}",
        f"options: {cfg.options}",
        f"export: {cfg.export}",
        f"version: {cfg.version}",
    ]
    return "; ".join(cfg_parts)


def update_image_metadata(
    image: ee.Image,
    interval_start: datetime,
    interval_end: datetime,
    cfg: ConfigParams,
) -> ee.Image:
    """Update the metadata of the given image with export configuration details.

    Args:
        image (ee.Image): The image whose metadata needs to be updated.
        interval_start (datetime): Start date of the interval.
        interval_end (datetime): End date of the interval.
        export_config (dict): The export configuration dictionary.

    Returns:
        ee.Image: The image with updated metadata.
        ["system:time_start", "system:time_end", "model", "variable", "export_scale", "output_crs", "system:index"]

        # system:index follows filenaming convention: [variable]_[model]_[output_bands (joined by '-')]_[output_resolution]m_s_[startdate(YYYYMMDD)]_[enddate(YYYYMMDD)]_[tile]_[crs(epsg lower and swap : with .)]_[version]

    """
    index_str = get_system_index(cfg, interval_start, interval_end)

    cfg_string = _get_cfg_string(cfg)

    updated_image = image.set(
        {
            "system:time_start": int(interval_start.timestamp() * 1000),
            "system:time_end": int(interval_end.timestamp() * 1000),
            "model": cfg.variables.model,
            "variable": cfg.variables.variable,
            "export_scale": cfg.export.scale,
            "output_crs": cfg.export.crs,
            "config": cfg_string,
            "system:index": index_str,
        },
    )
    return updated_image


def generate_intervals(start, end, temporal_interval):
    """Generate a list of (start, end) datetime tuples representing intervals
    between 'start' and 'end' based on the specified 'temporal_interval'.
    The 'temporal_interval' can be specified as:
    - An integer number of days
    - A timedelta object
    - A string representing common periods like "daily", "weekly", "monthly", etc.
    """
    logger.debug(
        f"Generating intervals from {start} to {end} with period {temporal_interval}",
    )
    if isinstance(temporal_interval, int):
        advance = lambda t: t + timedelta(days=temporal_interval)  # noqa
    elif isinstance(temporal_interval, str):
        temporal_interval = temporal_interval.lower()
        mapping = {
            "weekly": dict(weeks=1),
            "biweekly": dict(weeks=2),
            "monthly": dict(months=1),
            "bimonthly": dict(months=2),
            "quarterly": dict(months=3),
            "yearly": dict(years=1),
            "annual": dict(years=1),
        }
        if temporal_interval not in mapping:
            raise ValueError(f"Unknown period string: {temporal_interval}")
        delta = relativedelta(**mapping[temporal_interval])
        advance = lambda t: t + delta  # noqa
    else:
        raise TypeError("temporal_interval must be int, timedelta, or str")

    intervals = []
    current = start
    while current < end:
        nxt = advance(current)
        intervals.append((current, min(nxt, end)))
        current = nxt

    logger.debug(f"Generated {len(intervals)} time intervals.")
    return intervals


def asset_exists(asset_id: str) -> bool:
    """Check if a given Earth Engine asset exists.

    Args:
        asset_id (str): Full Earth Engine asset ID, e.g.,
                        "projects/ee-speckerfelix/assets/open-earth/lai_2023"

    Returns:
        bool: True if asset exists, False otherwise.

    """
    try:
        ee.data.getAsset(asset_id)
        return True
    except ee.EEException:
        return False


def set_asset_public(asset_id: str):
    """Sets a GEE asset to public read access."""
    acl = ee.data.getAssetAcl(asset_id)

    if "all_users_can_read" not in acl:
        # add 'all_users_can_read': False
        acl["all_users_can_read"] = True
        ee.data.setAssetAcl(asset_id, acl)
        logger.info(f"Set asset {asset_id} to public read access.")
    else:
        logger.info(f"Asset {asset_id} is already public.")


def initialize_export_location(cfg: ConfigParams, set_public: bool = False) -> None:
    loc = cfg.export.destination
    if loc == "asset":
        if asset_exists(cfg.export.collection_path):
            logger.warning(
                f"Asset {cfg.export.collection_path} already exists. Image exports with existing asset IDs will fail (delete beforehand if overwrite is desired).",
            )
        else:
            logger.info(
                f"Creating ImageCollection asset at {cfg.export.collection_path}.",
            )
            ee.data.createAsset({"type": "ImageCollection"}, cfg.export.collection_path)

            # add cfg string as property to ImageCollection asset
            cfg_string = _get_cfg_string(cfg)
            ee.data.setAssetProperties(
                cfg.export.collection_path,
                {"config": cfg_string},
            )

    if set_public:
        if cfg.export.destination != "asset":
            logger.warning(
                "Public access can only be set for asset exports. Skipping setting public access.",
            )
        else:
            logger.debug(
                "Note: The exported imageCollection will be set to public access, such that it can be visualized using the GEE-App: https://ee-speckerfelix.projects.earthengine.app/view/gee-biophys-export-visualizer"
            )
            set_asset_public(cfg.export.collection_path)

    elif loc == "drive":
        logger.info(
            f"Assets will be exported to Google Drive folder: {cfg.export.folder}",
        )

    elif loc == "gcs":
        logger.info(
            f"Assets will be exported to: gs://{cfg.export.bucket}/{cfg.export.folder}",
        )
    else:
        raise ValueError(f"Unknown output_location: {loc}")



import ee
import numpy as np
import pytest

from gee_biophys.models.sl2p import load_SL2P_model  # assumes you expose this

ATOL = 1e-6

# SL2P expects reflectances + cosines of angles (already cos-transformed)
BAND_ORDER_SL2P = [
    "cosSZA",
    "cosVZA",
    "cosRAA",
    "B3",
    "B4",
    "B5",
    "B6",
    "B7",
    "B8A",
    "B11",
    "B12",
]

N_SAMPLES = 10


@pytest.mark.parametrize("variable", ["lai", "fapar", "fcover"])
def test_sl2p_lai_random_inputs(ee_init, variable):
    rng = np.random.default_rng(42)

    # load paired models
    sl2p_mean_model, sl2p_std_model = load_SL2P_model(variable)

    for _ in range(N_SAMPLES):
        # --------- draw a valid sample ---------
        sample = {}
        # cosines are in [-1, 1]
        sample["cosSZA"] = float(rng.uniform(-1.0, 1.0))
        sample["cosVZA"] = float(rng.uniform(-1.0, 1.0))
        sample["cosRAA"] = float(rng.uniform(-1.0, 1.0))
        # surface reflectances in a typical range
        for b in ["B3", "B4", "B5", "B6", "B7", "B8A", "B11", "B12"]:
            sample[b] = float(rng.uniform(0.01, 0.5))

        # local predictions (sklearn)
        X = np.array([[sample[b] for b in BAND_ORDER_SL2P]], dtype=float)
        lai_mean_np = float(sl2p_mean_model.predict(X, clip_min_max=False).item())
        lai_std_np = float(sl2p_std_model.predict(X).item())

        # server-side predictions (single round-trip)
        img_in = ee.Image.constant([sample[b] for b in BAND_ORDER_SL2P]).rename(
            BAND_ORDER_SL2P
        )

        lai_mean_img = sl2p_mean_model.ee_predict(img_in).rename(f"{variable}_mean")
        lai_std_img = sl2p_std_model.ee_predict(img_in).rename(f"{variable}_std")
        out_img = ee.Image.cat([lai_mean_img, lai_std_img])

        # Sample once (constant image, but keep mask-safe pattern)
        pt = ee.Geometry.Point([0, 0])
        fc = out_img.sample(
            region=pt.buffer(50), scale=20, numPixels=1, dropNulls=True, tileScale=2
        )
        if fc.size().getInfo() == 0:
            vals = out_img.reduceRegion(
                reducer=ee.Reducer.first(),
                geometry=pt,
                scale=20,
                bestEffort=True,
                maxPixels=1e13,
            ).getInfo()
        else:
            vals = fc.first().toDictionary().getInfo()

        lai_mean_ee = float(vals[f"{variable}_mean"])
        lai_std_ee = float(vals[f"{variable}_std"])

        # assertions with good diagnostics
        try:
            np.testing.assert_allclose(lai_mean_np, lai_mean_ee, atol=ATOL)
            np.testing.assert_allclose(lai_std_np, lai_std_ee, atol=ATOL)
        except AssertionError as e:
            msg = (
                "\nSL2P LAI mismatch\n"
                f"Inputs (ordered {BAND_ORDER_SL2P}):\n"
                f"  {[sample[b] for b in BAND_ORDER_SL2P]}\n"
                f"Mean: local={lai_mean_np:.10f}, ee={lai_mean_ee:.10f}, Δ={lai_mean_np - lai_mean_ee:.2e}\n"
                f"Std : local={lai_std_np:.10f},  ee={lai_std_ee:.10f},  Δ={lai_std_np - lai_std_ee:.2e}\n"
            )
            raise AssertionError(msg) from e



from datetime import datetime
from typing import Literal

import ee
from loguru import logger

from gee_biophys.config import ConfigParams
from gee_biophys.models.s2biophys import prepare_s2_input_for_specker
from gee_biophys.models.sl2p import prepare_s2_input_for_sl2p


def get_s2_imgc(
    start_date: datetime,
    end_date: datetime,
    region: ee.Geometry,
    max_cloud_cover: int,
    bands: list = ["B2", "B3", "B4", "B5", "B6", "B7", "B8", "B8A", "B11", "B12"],
) -> ee.ImageCollection:
    """Retrieve Sentinel-2 image collection for the specified date range and region,
    selecting only the specified bands.
    """
    collection = (
        ee.ImageCollection("COPERNICUS/S2_HARMONIZED")
        .filterDate(start_date, end_date)
        .filterBounds(region)
        .filter(ee.Filter.lte("CLOUDY_PIXEL_PERCENTAGE", max_cloud_cover))
        .select(bands)
    )

    # divide DN values by 10000 to get reflectance, copy properties again
    def scale_reflectance(image):
        scaled = (
            image.select(bands)
            .divide(10000)
            .copyProperties(image, image.propertyNames())
        )
        return scaled

    collection = collection.map(scale_reflectance)
    return collection


def apply_cloudscore_plus_mask(
    s2_imgc: ee.ImageCollection,
    csplus_band: Literal["cs", "cs_cdf"],
    csplus_threshold: float,
) -> ee.ImageCollection:
    """Apply the CloudScorePlus algorithm to the given Sentinel-2 image collection
    and mask out cloudy pixels.
    """
    csplus_imgc = ee.ImageCollection("GOOGLE/CLOUD_SCORE_PLUS/V1/S2_HARMONIZED")

    linked_imgc = s2_imgc.linkCollection(csplus_imgc, [csplus_band])

    s2_imgc_masked = linked_imgc.map(
        lambda image: image.updateMask(image.select(csplus_band).gte(csplus_threshold)),
    )

    return s2_imgc_masked.select(s2_imgc.first().bandNames())


def add_angles_from_metadata_to_bands(image: ee.Image) -> ee.Image:
    """Add viewing/illumination angle bands (in degrees) derived from image metadata.

    This function reads Sentinel-2 metadata fields to:
      - compute the mean **view zenith** and **view azimuth** across bands B2,B3,B4,B5,B6,B7,B8,B8A,B11,B12,
      - read the **solar zenith** and **solar azimuth**,
      - and append three angle bands (no trigonometric transforms, no reflectance scaling):

        * tts — solar zenith angle (degrees)
        * tto — mean view zenith angle across the listed bands (degrees)
        * psi — absolute azimuth difference |view_azimuth − solar_azimuth| (degrees)

    Notes
    -----
    - Angles are kept in **degrees** (not cosine).

    - Expected metadata keys (Sentinel-2):
      `MEAN_SOLAR_AZIMUTH_ANGLE`, `MEAN_SOLAR_ZENITH_ANGLE`,
      `MEAN_INCIDENCE_AZIMUTH_ANGLE_<BAND>`, `MEAN_INCIDENCE_ZENITH_ANGLE_<BAND>`.

    Parameters
    ----------
    image : ee.Image
        Input image with the required Sentinel-2 angle metadata.

    Returns
    -------
    ee.Image
        The input image with added bands: 'tts', 'tto', and 'psi' (float32, degrees).

    """
    # Define the bands for which view angles are extracted from metadata.
    bands = ["B2", "B3", "B4", "B5", "B6", "B7", "B8", "B8A", "B11", "B12"]

    # Extract the solar azimuth and zenith angles from metadata.
    solar_azimuth = image.getNumber("MEAN_SOLAR_AZIMUTH_ANGLE")
    solar_zenith = image.getNumber("MEAN_SOLAR_ZENITH_ANGLE")

    # Calculate the mean view azimuth angle for the specified bands.
    view_azimuth = (
        ee.Array(
            [image.getNumber("MEAN_INCIDENCE_AZIMUTH_ANGLE_%s" % b) for b in bands],
        )
        .reduce(ee.Reducer.mean(), [0])
        .get([0])
    )

    # Calculate the mean view zenith angle for the specified bands.
    view_zenith = (
        ee.Array([image.getNumber("MEAN_INCIDENCE_ZENITH_ANGLE_%s" % b) for b in bands])
        .reduce(ee.Reducer.mean(), [0])
        .get([0])
    )

    # add tts, tto, psi
    image = image.addBands(ee.Image(solar_zenith).toFloat().rename("tts"))
    image = image.addBands(ee.Image(view_zenith).toFloat().rename("tto"))
    image = image.addBands(
        ee.Image(view_azimuth.subtract(solar_azimuth).abs()).toFloat().rename("psi"),
    )

    return image


def load_s2_input(
    cfg: ConfigParams,
    interval_start: datetime,
    interval_end: datetime,
) -> ee.ImageCollection:
    """Load and prepare Sentinel-2 ImageCollection based on configuration parameters."""
    s2_imgc = get_s2_imgc(
        start_date=interval_start,
        end_date=interval_end,
        region=cfg.spatial.ee_geometry,
        max_cloud_cover=cfg.options.max_cloud_cover,
    )
    logger.debug(f"Input S2 ImageCollection size: {s2_imgc.size().getInfo()}")

    s2_imgc = apply_cloudscore_plus_mask(
        s2_imgc,
        csplus_band=cfg.options.csplus_band,
        csplus_threshold=cfg.options.cs_plus_threshold,
    )

    # Add angles from metadata to bands
    s2_imgc = s2_imgc.map(add_angles_from_metadata_to_bands)

    if cfg.variables.model == "sl2p":
        s2_imgc = s2_imgc.map(prepare_s2_input_for_sl2p)
    elif cfg.variables.model == "s2biophys":
        s2_imgc = s2_imgc.map(prepare_s2_input_for_specker)
    else:
        raise ValueError(f"Unsupported model '{cfg.variables.model}'")

    return s2_imgc


# if __name__ == "__main__":
#     ee.Initialize()

#     start = datetime(2023, 6, 1)
#     end = datetime(2023, 6, 30)
#     region = ee.Geometry.Point([-122.262, 37.8719]).buffer(10000)  # 10 km buffer

#     bands = ["B4", "B3", "B2"]  # Red, Green, Blue bands

#     s2_imgc = get_s2_imgc(start, end, region, bands, max_cloud_cover=50)
#     print(f"Original ImageCollection size: {s2_imgc.size().getInfo()}")

#     s2_imgc_masked = apply_cloudscore_plus_mask(
#         s2_imgc, csplus_band="cs", csplus_threshold=0.7
#     )
#     print(f"Masked ImageCollection size: {s2_imgc_masked.size().getInfo()}")



import ee
from loguru import logger

from gee_biophys.config import ConfigParams


def export_image(image: ee.Image, filename: str, cfg: ConfigParams):
    loc = cfg.export.destination

    if cfg.spatial.type == "geojson" and cfg.spatial.geojson_clip:
        logger.debug(
            "Clipping export image to GeoJSON geometry bounds. Please be aware of potential issues with complex geometries. Set 'geojson_clip' to false to disable.",
        )
        image = image.clip(cfg.spatial.ee_geometry)

    if loc == "asset":
        asset_id = f"{cfg.export.collection_path}/{filename}"
        logger.debug(f"Exporting image to Asset: {asset_id}")
        task = ee.batch.Export.image.toAsset(
            image=image,
            description=filename,
            assetId=asset_id,
            region=cfg.spatial.ee_geometry,
            scale=cfg.export.scale,
            crs=cfg.export.crs,
            maxPixels=cfg.export.max_pixels,
        )
        task.start()
    elif loc == "drive":
        image = image.float()  # to avoid error: Exported bands must have compatible data types; found inconsistent types: Float64 and Int32
        asset_id = f"{cfg.export.folder}/{filename}"
        logger.debug(
            f"Exporting image to Google Drive file: {cfg.export.folder}/{filename}",
        )
        task = ee.batch.Export.image.toDrive(
            image=image,
            description=filename,
            folder=cfg.export.folder,
            fileNamePrefix=filename,
            region=cfg.spatial.ee_geometry,
            scale=cfg.export.scale,
            crs=cfg.export.crs,
            maxPixels=1e11,
        )
        task.start()
    elif loc == "gcs":
        image = image.float()  # to avoid error: Exported bands must have compatible data types; found inconsistent types: Float64 and Int32
        bucket = cfg.export.bucket
        gcs_folder = cfg.export.folder

        logger.debug(
            f"Exporting image to GCS bucket: gs://{bucket}/{gcs_folder}/{filename}",
        )
        task = ee.batch.Export.image.toCloudStorage(
            image=image,
            description=filename,
            bucket=bucket,
            fileNamePrefix=f"{gcs_folder}/{filename}" if gcs_folder else filename,
            region=cfg.spatial.ee_geometry,
            scale=cfg.export.scale,
            crs=cfg.export.crs,
            maxPixels=1e11,
        )
        task.start()
    else:
        raise ValueError(f"Unknown output_location: {loc}")

    logger.debug(f"Task started with ID: {task.id}")



import json
from importlib.resources import files
from typing import Literal, Optional

import ee
import numpy as np
from loguru import logger


class LeafToolbox_MLPRegressor:
    def __init__(
        self,
        net: dict,
        domain_codes: list = None,
    ) -> None:
        ### init original params
        # load params from json dict
        self.inp_slope = np.array(net["inp_slope"])
        self.inp_offset = np.array(net["inp_offset"])
        self.h1wt = np.array(net["h1wt"]).reshape(
            len(net["h1bi"]),
            len(self.inp_offset),
        )
        self.h1bi = np.array(net["h1bi"])
        self.h2wt = np.array(net["h2wt"]).reshape(1, len(self.h1bi))
        self.h2bi = np.array(net["h2bi"])
        self.out_slope = np.array(net["out_slope"])
        self.out_bias = np.array(net["out_bias"])
        self.bandorder = net["bandorder"]

        ### init ee params
        self.ee_inp_slope = ee.Array(self.inp_slope.tolist())
        self.ee_inp_offset = ee.Array(self.inp_offset.tolist())
        self.ee_h1wt = ee.Array(
            self.h1wt.tolist(),
        ).transpose()  # its crucial to use transpose instead of reshape here!! (otherwise values fit into wrong positions)

        self.ee_h1bi = ee.Array(self.h1bi.tolist()).reshape([1, -1])
        self.ee_h2wt = ee.Array(self.h2wt.tolist()).transpose()

        self.ee_h2bi = ee.Array(self.h2bi.tolist()).reshape([1, -1])
        self.ee_out_slope = ee.Array(self.out_slope.tolist())
        self.ee_out_bias = ee.Array(self.out_bias.tolist())

        logger.debug(
            f"SL2P init(): Make sure that the input data is ordered as in bandorder: {self.bandorder}",
        )

        if domain_codes is not None:
            self.domain_codes = domain_codes
            logger.debug(
                f"SL2P init(): Domain codes for input validation provided: {self.domain_codes}",
            )
            self.init_domain_codes()
        else:
            self.domain_codes = None
            logger.debug("SL2P init(): No domain codes for input validation provided")

    def _tansig(self, x):
        return 2.0 / (1.0 + np.exp(-2.0 * x)) - 1.0

    def init_domain_codes(self):
        raise NotImplementedError("Domain code initialization not implemented yet.")

    def check_domain(self, X: np.ndarray) -> np.ndarray:
        """X: (n_samples, n_features)
        returns: boolean array of shape (n_samples,) indicating whether each sample is valid
        """
        if self.domain_codes is None:
            raise ValueError("Domain codes not initialized.")

        # Image formatting
        image_format = np.sum(
            (np.uint8(np.ceil(X * 10) % 10))
            * np.array([10**value for value in range(len(self.bandorder))])[:, None],
            axis=0,
        )

        # Comparing image to domain codes
        flag = np.isin(image_format, self.domain_codes, invert=True)
        return flag

    def ee_check_domain(self, img: ee.Image) -> ee.Image:
        """img: ee.Image with bands ordered as in self.bandorder
        returns: ee.Image with 1 for invalid pixels, 0 for valid pixels
        """
        # TODO: check function carefully!!!!!!! - Copilot generated
        # Add test case that check_domain() and ee_check_domain() give same results on same data

        if self.domain_codes is None:
            raise ValueError("Domain codes not initialized.")

        # Image formatting
        image_format = ee.Image.constant(0)
        for i, band in enumerate(self.bandorder):
            band_component = (
                img.select(band).multiply(10).mod(10).floor().multiply(10**i)
            )
            image_format = image_format.add(band_component)

        # Comparing image to domain codes
        domain_codes_ee = ee.List(self.domain_codes)
        invalid_mask = image_format.remap(
            domain_codes_ee,
            ee.List.repeat(0, domain_codes_ee.size()),
            1,
        )
        return invalid_mask

    def predict(
        self,
        X: np.ndarray,
        clip_min_max: bool = False,
        clip_variable: Optional[str] = None,
    ) -> np.ndarray:
        """X: (n_samples, n_features)
        returns: (n_samples,)
        """
        # input scaling
        Xs = X * self.inp_slope + self.inp_offset

        # hidden layer
        h1 = self._tansig(np.matmul(Xs, self.h1wt.T) + self.h1bi)

        # linear output layer
        h2 = np.matmul(h1, self.h2wt.T) + self.h2bi

        # output scaling
        y = (h2 - self.out_bias) / self.out_slope

        if clip_min_max:
            assert clip_variable is not None, (
                "clip_variable must be provided when clip_min_max is True"
            )
            if clip_variable in ["lai", "laie"]:
                y = np.clip(y, 0, 8)
            elif clip_variable in ["fapar", "fcover"]:
                y = np.clip(y, 0, 1)
            else:
                raise ValueError(f"Clipping not defined for variable: {clip_variable}")
        return y.ravel()

    def _tansig_ee(self, x: ee.Image) -> ee.Image:
        return (
            ee.Image(2)
            .divide(ee.Image(1).add(x.multiply(ee.Image(-2)).exp()))
            .subtract(ee.Image(1))
        )

    def ee_predict(
        self,
        ee_img: ee.Image,
        clip_min_max: bool = False,
        clip_variable: Optional[str] = None,
    ) -> ee.Image:
        """Predict using an ee.Image as input.
        Returns an ee.Image with the prediction.
        """
        # TODO: check function carefully!!!!!!! - Copilot generated
        # Add test case that predict() and ee_predict() give same results on same data

        x = ee_img.toArray()
        x = x.multiply(ee.Image(self.ee_inp_slope)).add(ee.Image(self.ee_inp_offset))

        # convert to 2D array for matrix multiplication
        x = x.toArray(1).arrayTranspose()  # shape : (1, bands)

        h1 = self._tansig_ee(
            x.matrixMultiply(ee.Image(self.ee_h1wt)).add(ee.Image(self.ee_h1bi)),
        )  # (1, bands) x (bands, n_nodes) = (1, n_nodes)

        # linear output layer
        h2 = h1.matrixMultiply(ee.Image(self.ee_h2wt)).add(ee.Image(self.ee_h2bi))
        # (1, n_nodes) x (n_nodes, 1) = (1, 1)

        y = (
            h2.arrayProject([0])
            .subtract(ee.Image(self.ee_out_bias))
            .divide(ee.Image(self.ee_out_slope))
        )

        if clip_min_max:
            assert clip_variable is not None, (
                "clip_variable must be provided when clip_min_max is True"
            )
            if clip_variable == "lai":
                y = y.clamp(0, 8)
            elif clip_variable in ["fapar", "fcover"]:
                y = y.clamp(0, 1)

        return y.arrayFlatten([["output"]])

    def predict_with_domain_check(self, X: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        if self.domain_codes is None:
            raise ValueError("Domain codes not initialized.")

        invalid_mask = self.check_domain(X)
        if np.any(invalid_mask):
            logger.warning(
                f"{np.sum(invalid_mask)} samples are outside the valid input domain.",
            )

        return self.predict(X), invalid_mask


def load_SL2P_model(
    variable: Literal["lai", "laie", "fapar", "fcover"],
) -> tuple[LeafToolbox_MLPRegressor, LeafToolbox_MLPRegressor]:
    # remap to trait names used in SL2P parameter files
    trait_map = {
        "lai": "LAI",
        "laie": "LAI",
        "fapar": "fAPAR",
        "fcover": "fCOVER",
    }
    variable_mapped = trait_map[variable]

    package = "gee_biophys.models.sl2p_params"

    estimate_file = f"{variable_mapped}-estimation_SL2P_Corrected_LeafToolBox.json"
    uncertainty_json = f"{variable_mapped}-uncertainty_SL2P_Corrected_LeafToolBox.json"

    with (files(package) / estimate_file).open("r") as f:
        estimate_params = json.load(f)

    with (files(package) / uncertainty_json).open("r") as f:
        uncertainty_params = json.load(f)

    model_estimate = LeafToolbox_MLPRegressor(estimate_params, domain_codes=None)
    model_uncertainty = LeafToolbox_MLPRegressor(uncertainty_params, domain_codes=None)

    return (
        model_estimate,
        model_uncertainty,
    )


def _ee_angle_transform_sl2p(angle_img: ee.Image) -> ee.Image:
    """Transform angle bands from degrees to the format required by SL2P:
    - Convert degrees to radians
    - Compute cosine of the angles

    Parameters
    ----------
    - angle_img (ee.Image): Image with angle bands in degrees.

    Returns
    -------
    - ee.Image: Image with angle bands transformed for SL2P.

    """
    radians_img = angle_img.multiply(np.pi / 180.0)
    cos_img = radians_img.cos()
    return cos_img


def prepare_s2_input_for_sl2p(img: ee.Image) -> ee.Image:
    """Prepare Sentinel-2 image for SL2P model input.

    Parameters
    ----------
    - img (ee.Image): Input Sentinel-2 image with bands and angles.

    Returns
    -------
    - ee.Image: Image with bands ordered and angles transformed for SL2P.

    """
    # Bands/angles expected on the input S2 image
    _s2_bands = ["B3", "B4", "B5", "B6", "B7", "B8A", "B11", "B12"]
    _s2_angles = ["tts", "tto", "psi"]

    # Output names required by SL2P (cosines first, then reflectance bands)
    _sl2p_angle_names = ["cosSZA", "cosVZA", "cosRAA"]
    _sl2p_output_order = _sl2p_angle_names + _s2_bands

    refl = img.select(_s2_bands)
    cos_angles = _ee_angle_transform_sl2p(img.select(_s2_angles)).rename(
        _sl2p_angle_names,
    )

    out = refl.addBands(cos_angles)

    return out.select(_sl2p_output_order)



from datetime import UTC, datetime


from gee_biophys.config import Temporal


def test_intervals():
    """Test correct interval generation for various cadences / fixed, seasonal."""

    # test valid monthly and bimonthly cadences
    data = {
        "start": "2023-01-01",
        "end": "2023-12-31",
        "cadence": {"type": "fixed", "interval": "monthly"},
    }
    t = Temporal(**data)

    assert t.start == datetime(2023, 1, 1, 0, 0, 0, tzinfo=UTC)
    assert t.end == datetime(2023, 12, 31, 0, 0, 0, tzinfo=UTC)

    assert len(list(t.iter_date_ranges())) == 12

    data = {
        "start": "2023-01-15",
        "end": "2023-04-14",
        "cadence": {"type": "fixed", "interval": "bimonthly"},
    }
    t = Temporal(**data)
    assert len(list(t.iter_date_ranges())) == 2

    # test valid seasonal cadence - in-year
    data = {
        "start": "2021-01-01",
        "end": "2023-12-31",
        "cadence": {"type": "seasons", "start": "05-15", "end": "09-15"},
    }
    t = Temporal(**data)
    assert len(list(t.iter_date_ranges())) == 3
    ranges = list(t.iter_date_ranges())
    assert ranges[0] == (
        datetime(2021, 5, 15, 0, 0, 0, tzinfo=UTC),
        datetime(2021, 9, 15, 0, 0, 0, tzinfo=UTC),
    )
    assert ranges[1] == (
        datetime(2022, 5, 15, 0, 0, 0, tzinfo=UTC),
        datetime(2022, 9, 15, 0, 0, 0, tzinfo=UTC),
    )
    assert ranges[2] == (
        datetime(2023, 5, 15, 0, 0, 0, tzinfo=UTC),
        datetime(2023, 9, 15, 0, 0, 0, tzinfo=UTC),
    )

    # test valid seasonal cadence - cross-year
    data = {
        "start": "2020-01-01",
        "end": "2022-12-31",
        "cadence": {"type": "seasons", "start": "11-01", "end": "03-01"},
    }
    # this should raise warning, because start is within season
    t = Temporal(**data)
    assert len(list(t.iter_date_ranges())) == 4
    ranges = list(t.iter_date_ranges())
    assert ranges[0] == (
        datetime(2020, 1, 1, 0, 0, 0, tzinfo=UTC),
        datetime(2020, 3, 1, 0, 0, 0, tzinfo=UTC),
    )
    assert ranges[1] == (
        datetime(2020, 11, 1, 0, 0, 0, tzinfo=UTC),
        datetime(2021, 3, 1, 0, 0, 0, tzinfo=UTC),
    )
    assert ranges[2] == (
        datetime(2021, 11, 1, 0, 0, 0, tzinfo=UTC),
        datetime(2022, 3, 1, 0, 0, 0, tzinfo=UTC),
    )
    assert ranges[3] == (
        datetime(2022, 11, 1, 0, 0, 0, tzinfo=UTC),
        datetime(2022, 12, 31, 0, 0, 0, tzinfo=UTC),
    )



import pytest
import yaml

from gee_biophys.config import ConfigParams
from gee_biophys.s2_input import load_s2_input
from gee_biophys.s2_predict import biophys_predict

ATOL = 1e-5
N_SAMPLES = 10


def load_params(path: str) -> ConfigParams:
    with open(path) as f:
        data = yaml.safe_load(f)
    return ConfigParams(**data)


@pytest.mark.parametrize(
    "config_path",
    [
        "example_configs/minimal_example.yaml",
        "example_configs/bimonthly-zambia.yaml",
        "example_configs/seasonal-summer-zurich.yaml",
    ],
)
def test_cli(ee_init, config_path):
    cfg = load_params(config_path)

    for interval_start, interval_end in cfg.temporal.iter_date_ranges():
        imgc = load_s2_input(cfg, interval_start, interval_end)
        imgc.getInfo()  # force evaluation to catch errors

        output_image = biophys_predict(cfg, imgc)
        output_image.getInfo()  # force evaluation to catch errors
        break


@pytest.mark.parametrize(
    "config_path",
    [
        "example_configs/minimal_example.yaml",
        "example_configs/bimonthly-zambia.yaml",
        "example_configs/seasonal-summer-zurich.yaml",
        "example_configs/forest-fire-bitsch-2023.yaml",
    ],
)
def test_configs(config_path):
    cfg = load_params(config_path)

    if config_path == "example_configs/forest-fire-bitsch-2023.yaml":
        assert cfg.export.crs == "EPSG:32632"  # ensure LOCAL_UTM is resolved

    assert isinstance(cfg, ConfigParams)



import json
from dataclasses import dataclass
from importlib.resources import files
from pathlib import Path
from pickle import load as pickle_load
from typing import Any

import ee
import numpy as np
from sklearn.pipeline import Pipeline

from gee_biophys.models.utils_s2biophys import (
    eeMinMaxRangeMasker,
    eeMLPRegressor,
    eeStandardScaler,
)


def ee_nirv_normalisation(image: ee.Image):
    reflectance_bands = ["B2", "B3", "B4", "B5", "B6", "B7", "B8", "B8A", "B11", "B12"]
    NDVI = image.normalizedDifference(["B8", "B4"]).rename("NDVI")
    NIRv = NDVI.multiply(image.select("B8")).rename("NIRv")

    image_normalized = image.select(reflectance_bands).divide(NIRv)
    image_to_return = image.addBands(image_normalized, overwrite=True)
    return image_to_return


def ee_angle_transformer(image: ee.Image):
    # cosine transformation of angles
    image_angles = image.select(["tts", "tto", "psi"]).multiply(np.pi / 180).cos()
    image_to_return = image.addBands(image_angles, overwrite=True)
    return image_to_return


def ee_logit_transform(image: ee.Image, trait: str):
    # logit transformation of trait
    #  np.log(x / (1 - x))
    image = image.addBands(
        image.select(trait).log().divide(image.select(trait).subtract(1)),
    )
    return image


def ee_logit_inverse_transform(image: ee.Image, trait: str):
    # inverse logit transformation of trait
    # 1 / (1 + np.exp(-x))
    return (
        image.select(trait).expression("1 / (1 + exp(-x))", {"x": image}).rename(trait)
    )


def ee_log1p_inverse_transform(image: ee.Image, trait: str):
    # inverse log1p transformation of trait
    # np.exp(x) - 1
    return image.select(trait).exp().subtract(1).rename(trait)


def eePipelinePredictMap(
    pipeline: Pipeline,
    imgc: ee.ImageCollection,
    trait: str,
    model_config: dict,
    clip_min_max: bool,
    min_max_bands: dict | None = None,
):
    # get the bands and angles
    bands = ["B2", "B3", "B4", "B5", "B6", "B7", "B8", "B8A", "B11", "B12"]
    angles = ["tts", "tto", "psi"]

    # mask all pixels with reflectance values outside of the min_max reflectance values
    if min_max_bands is not None:
        min_max_band_masker = eeMinMaxRangeMasker(min_max_bands)
        imgc = imgc.map(min_max_band_masker.ee_mask)

    if model_config["nirv_norm"]:
        imgc = imgc.map(ee_nirv_normalisation)

    features = bands + angles
    imgc = imgc.map(ee_angle_transformer)
    imgc = imgc.select(features)

    # always apply standard scaler
    band_scaler = (
        pipeline.named_steps["preprocessor"]
        .named_transformers_["band_transformer"]
        .named_steps["scaler"]
    )

    ee_band_scaler = eeStandardScaler(band_scaler)
    # a = ee_band_scaler.transform_image(imgc.first())
    imgc = imgc.map(ee_band_scaler.transform_image)

    # apply model:
    if model_config["model"] == "mlp":
        # IMPORTANT: .regressor_ refers to the actual model, while .regressor only refers to the untrained model
        ee_model = eeMLPRegressor(
            pipeline.named_steps["regressor"].regressor_,
            trait_name=trait,
        )
    else:
        raise ValueError("Only mlp models are supported for now")
    imgc = imgc.map(lambda image: ee_model.predict(image))

    # apply inverse transformations
    if model_config["transform_target"] == "log1p":
        imgc = imgc.map(
            lambda image: ee_log1p_inverse_transform(image, trait).copyProperties(
                image
            ),
        )
    elif model_config["transform_target"] == "logit":
        imgc = imgc.map(
            lambda image: ee_logit_inverse_transform(image, trait).copyProperties(
                image
            ),
        )
    elif model_config["transform_target"] == "standard":
        target_scaler = pipeline.named_steps["regressor"].transformer_
        target_ee_scaler = eeStandardScaler(
            target_scaler,
            feature_names=[trait],
        )  # must be a list
        imgc = imgc.map(
            lambda image: target_ee_scaler.inverse_transform_column(
                image,
                trait,
            ).copyProperties(image),
        )
    elif model_config["transform_target"] == "None":
        imgc = imgc
    else:
        raise ValueError(
            f"Unknown target transformation: {model_config['transform_target']}",
        )

    if clip_min_max:
        clip_dict = {
            "lai": (0, 8),
            "laie": (0, 8),
            "fapar": (0, 1),
            "fcover": (0, 1),
        }
        if trait in clip_dict:
            min_val, max_val = clip_dict[trait]
            imgc = imgc.map(
                lambda image: image.clamp(min_val, max_val).copyProperties(image),
            )
        else:
            raise ValueError(f"Clipping not defined for trait: {trait}")

    return imgc


@dataclass
class EnsembleItem:
    config: dict
    pipeline: Any
    model_path: str
    min_max_bands: dict
    min_max_label: dict
    split: dict

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)


def _open_json(p: Path) -> dict:
    with p.open("r", encoding="utf-8") as f:
        return json.load(f)


def _open_pickle(p: Path):
    # If you have a custom safe loader, swap it here.
    with p.open("rb") as f:
        return pickle_load(f)


def load_model_ensemble(trait: str) -> dict:
    base = files("gee_biophys.models.specker_params") / trait
    if not base.is_dir():
        raise FileNotFoundError(
            f"Packaged models for trait '{trait}' not found at {base}. "
            "Ensure files are included via [tool.setuptools.package-data].",
        )

    n_testsets = 5
    models: dict[str, EnsembleItem] = {}

    for t in range(n_testsets):
        name = f"optuna-v2-{trait}-mlp-split-{t}"

        pipeline_path = base / f"model_{name}.pkl"
        config_path = base / f"model_{name}_config.json"
        min_max_bands_path = base / f"min_max_band_values_{name}.json"
        min_max_label_path = base / f"min_max_label_values_{name}.json"
        split_path = base / f"model_{name}_split.json"

        required = {
            "pipeline": pipeline_path,
            "config": config_path,
            "min_max_bands": min_max_bands_path,
            "min_max_label": min_max_label_path,
            "split": split_path,
        }

        missing = [k for k, p in required.items() if not p.is_file()]
        if missing:
            # Give a helpful, actionable error
            details = "\n".join(f"  - {k}: {required[k]}" for k in missing)
            raise FileNotFoundError(
                f"Missing required model files for '{name}':\n{details}\n"
                "Make sure they are packaged and names match the expected pattern.",
            )

        item = EnsembleItem(
            config=_open_json(config_path),
            pipeline=_open_pickle(pipeline_path),
            model_path=pipeline_path.with_suffix("").name,
            min_max_bands=_open_json(min_max_bands_path),
            min_max_label=_open_json(min_max_label_path),
            split=_open_json(split_path),
        )
        models[name] = item

    return models


def prepare_s2_input_for_specker(img: ee.Image) -> ee.Image:
    """Prepare Sentinel-2 image for Specker et al. model prediction by selecting and ordering bands.

    Parameters
    ----------
    - img (ee.Image): Input Sentinel-2 image with bands.

    Returns
    -------
    - ee.Image: Image with bands ordered as required by Specker et al. model.

    """
    bands = [
        "B2",
        "B3",
        "B4",
        "B5",
        "B6",
        "B7",
        "B8",
        "B8A",
        "B11",
        "B12",
    ]
    angles = ["tts", "tto", "psi"]
    # reorder bands
    band_order_reorder = bands + angles

    return img.select(band_order_reorder)



import json
import math
import re
from collections.abc import Iterator
from datetime import UTC, datetime, timedelta
from functools import cached_property
from pathlib import Path
from typing import Annotated, Literal

import ee
from loguru import logger
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PositiveInt,
    field_validator,
    model_validator,
)


# ----------------- Spatial -----------------
class Spatial(BaseModel):
    type: Literal["bbox", "geojson"]
    bbox: list[float] | None = None  # [minx, miny, maxx, maxy]
    geojson_path: str | None = None
    region_name: str | None = (
        None  # optional name for the region (no spaces/underscores)
    )
    geojson_clip: bool = Field(
        default=False,
        description="If true and type is 'geojson', clip to geometry bounds.",
    )

    # Enforce that *no other keys* are accepted
    model_config = ConfigDict(extra="forbid")

    @model_validator(mode="after")
    def check_inputs(self):
        has_bbox = self.bbox is not None
        has_geo = self.geojson_path is not None
        has_region_name = self.region_name is not None

        if has_region_name:
            # assert not spaces and underscores
            if re.search(r"[ _]", self.region_name):
                raise ValueError(
                    "spatial.region_name must not contain spaces or underscores.",
                )

        # exactly one must be provided
        if has_bbox == has_geo:
            raise ValueError(
                "Specify exactly one of 'bbox' or 'geojson_path', not both.",
            )

        if self.type == "bbox":
            if not has_bbox:
                raise ValueError("When type='bbox', 'bbox' is required.")
            if len(self.bbox) != 4:
                raise ValueError(
                    "bbox must contain four numbers: [minx, miny, maxx, maxy].",
                )
            minx, miny, maxx, maxy = self.bbox
            if not (minx < maxx and miny < maxy):
                raise ValueError("bbox must satisfy minx < maxx and miny < maxy.")
            # optional lon/lat sanity
            if not (
                -180 <= minx <= 180
                and -180 <= maxx <= 180
                and -90 <= miny <= 90
                and -90 <= maxy <= 90
            ):
                raise ValueError("bbox coordinates must be within lon/lat bounds.")
        elif self.type == "geojson":
            if not has_geo:
                raise ValueError("When type='geojson', 'geojson_path' is required.")
            p = Path(self.geojson_path)
            if not p.exists():
                raise ValueError(f"geojson_path does not exist: {p}")
            if p.suffix.lower() not in {".geojson", ".json"}:
                raise ValueError("geojson_path must end with .geojson or .json.")
            # quick JSON check
            json.loads(p.read_text(encoding="utf-8"))

        return self

    @cached_property
    def ee_geometry(self) -> ee.Geometry:
        """Return an ee.Geometry derived from this spatial definition."""
        if self.type == "bbox":
            return ee.Geometry.BBox(*self.bbox)
        if self.type == "geojson":
            with open(self.geojson_path, encoding="utf-8") as f:
                obj = json.load(f)

            t = obj.get("type", "").lower()

            # type needs to be: polygon, multipolygon, feature, featurecollection (everything else is rejected)
            if t not in (
                "polygon",
                "multipolygon",
                "feature",
                "featurecollection",
            ):
                raise ValueError(
                    f"GeoJSON 'type' must be 'Polygon', 'MultiPolygon', 'Feature', or 'FeatureCollection', got '{t}'.",
                )

            if t in ("polygon", "multipolygon"):
                # Raw Geometry dict
                return ee.Geometry(obj)

            if t == "feature":
                # Single Feature -> use its geometry
                geom = obj.get("geometry")
                if not geom:
                    raise ValueError("GeoJSON Feature has no 'geometry'.")
                return ee.Geometry(geom)

            if t == "featurecollection":
                # Full FeatureCollection -> dissolve to one geometry
                fc = ee.FeatureCollection(obj)
                if fc.size().getInfo() == 1:
                    # single feature -> use its geometry
                    first = ee.Feature(fc.first())
                    return first.geometry()
                logger.warning(
                    "GeoJSON FeatureCollection has multiple features; dissolving to single geometry.",
                )
                return fc.geometry()

            raise ValueError(
                f"Unsupported or missing GeoJSON 'type': {obj.get('type')}",
            )

        raise ValueError(f"Unsupported spatial type: {self.type}")


_MM_DD = re.compile(r"^(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01])$")


def _safe_ymd(year: int, month: int, day: int) -> datetime:
    """Return a valid UTC datetime, replacing Feb 29 -> Feb 28 if needed."""
    try:
        return datetime(year, month, day, tzinfo=UTC)
    except ValueError:
        if month == 2 and day == 29:
            return datetime(year, 2, 28, tzinfo=UTC)
        raise


# ---------- Cadences ----------
class FixedCadence(BaseModel):
    type: Literal["fixed"]
    interval: (
        int
        | Literal[
            "weekly",
            "biweekly",
            "monthly",
            "bimonthly",
            "quarterly",
            "yearly",
            "annual",
        ]
    )

    model_config = ConfigDict(extra="forbid")

    @field_validator("interval", mode="before")
    def _normalize_interval(cls, v):
        if isinstance(v, int):
            if v <= 0:
                raise ValueError("interval (int days) must be > 0.")
            return v
        if isinstance(v, str):
            s = v.strip().lower()
            return "yearly" if s == "annual" else s
        raise ValueError("interval must be int days or a valid period string.")


class SeasonsCadence(BaseModel):
    type: Literal["seasons"]
    start: str  # "MM-DD"
    end: str  # "MM-DD"
    model_config = ConfigDict(extra="forbid")

    @field_validator("start", "end")
    def _mmdd(cls, v):
        if not _MM_DD.match(v):
            raise ValueError("Use zero-padded 'MM-DD'.")
        return v


Cadence = Annotated[FixedCadence | SeasonsCadence, Field(discriminator="type")]


# ----------------- Temporal -----------------
class Temporal(BaseModel):
    # store as aware UTC datetimes; accept ISO strings in JSON
    start: datetime
    end: datetime
    cadence: Cadence

    # Enforce that *no other keys* are accepted
    model_config = ConfigDict(extra="forbid")

    @field_validator("start", "end", mode="before")
    def _parse_iso(cls, v):
        if not isinstance(v, str):
            raise ValueError("Dates must be strings in ISO 8601 format.")
        s = v.strip()
        try:
            # support 'YYYY-MM-DD' and 'YYYY-MM-DDTHH:MM:SS[.fff][Z|+00:00]'
            if s.endswith("Z"):
                dt = datetime.fromisoformat(s.replace("Z", "+00:00"))
            else:
                dt = datetime.fromisoformat(s)
            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=UTC)
            if dt.utcoffset() != timedelta(0):
                raise ValueError("Datetime must be UTC (use 'Z' or +00:00).")
            return dt
        except Exception:
            raise ValueError(
                f"Invalid date string '{v}'. Use 'YYYY-MM-DD' or 'YYYY-MM-DDTHH:MM:SSZ'.",
            )

    # Public API: yield (start_dt, end_dt)
    def iter_date_ranges(self) -> Iterator[tuple[datetime, datetime]]:
        if isinstance(self.cadence, FixedCadence):
            yield from self._iter_fixed(self.cadence)
        else:
            yield from self._iter_seasons(self.cadence)

    # ---- Fixed intervals ----
    def _iter_fixed(self, c: FixedCadence) -> Iterator[tuple[datetime, datetime]]:
        """Yield fixed-length intervals clipped to [self.start, self.end).
        Raise a warning if the final interval is truncated at the temporal end.
        """
        if isinstance(c.interval, int):
            step = timedelta(days=c.interval)
            t0 = self.start
            while t0 < self.end:
                t1 = min(t0 + step, self.end)
                if t1 != t0 + step:
                    logger.warning(
                        f"Final interval from {t0.date()} to {t1.date()} is truncated "
                        f"to fit within temporal.end {self.end.date()}.",
                    )
                if t1 > t0:
                    yield (t0, t1)
                t0 = t1
            return

        name = c.interval
        if name in {"weekly", "biweekly"}:
            step = timedelta(days=7 if name == "weekly" else 14)
            t0 = self.start
            while t0 < self.end:
                t1 = min(t0 + step, self.end)
                if t1 != t0 + step:
                    logger.warning(
                        f"Final interval '{name}' from {t0.date()} to {t1.date()} is truncated "
                        f"to fit within temporal.end {self.end.date()}.",
                    )
                if t1 > t0:
                    yield (t0, t1)
                t0 = t1
            return

        months = {"monthly": 1, "bimonthly": 2, "quarterly": 3, "yearly": 12}[name]
        t = self.start
        while t < self.end:
            y, m = t.year, t.month
            new_m = m + months
            y2 = y + (new_m - 1) // 12
            m2 = (new_m - 1) % 12 + 1
            end_block = datetime(y2, m2, 1, tzinfo=UTC)
            t1 = min(end_block, self.end)
            if t1 != end_block:
                logger.warning(
                    f"Final interval '{name}' from {t.date()} to {t1.date()} is truncated "
                    f"to fit within temporal.end {self.end.date()}.",
                )
            if t1 > t:
                yield (t, t1)
            t = t1

    # ---- Seasonal intervals (cross-year allowed) ----
    def _iter_seasons(self, c: SeasonsCadence) -> Iterator[tuple[datetime, datetime]]:
        """Yield exactly ONE tuple per season-year:
        - same-year:  YYYY-MM-DD -> YYYY-MM-DD
        - cross-year: YYYY-MM-DD -> (YYYY+1)-MM-DD
        All windows are clipped to [self.start, self.end).
        """
        m1, d1 = map(int, c.start.split("-"))
        m2, d2 = map(int, c.end.split("-"))

        y = self.start.year - 1  # include prior year to catch wrap crossing self.start
        last_y = self.end.year

        while y <= last_y:
            a = _safe_ymd(y, m1, d1)  # nominal start in year y
            b_same = _safe_ymd(y, m2, d2)  # same-year end if not wrapping
            win_start = a
            win_end = b_same if a < b_same else _safe_ymd(y + 1, m2, d2)

            # clip to global [start, end)
            start_clipped = max(win_start, self.start)
            end_clipped = min(win_end, self.end)

            # warn if clipped at either boundary
            if start_clipped < end_clipped and (
                start_clipped > win_start or end_clipped < win_end
            ):
                logger.warning(
                    f"Season ({c.start} until {c.end}) around {y} is truncated by "
                    f"[{self.start.date()} — {self.end.date()}].",
                )

            if start_clipped < end_clipped:
                yield (start_clipped, end_clipped)

            y += 1

    @model_validator(mode="after")
    def _check_order_and_warn_truncation(self):
        if self.start >= self.end:
            raise ValueError("temporal.start must be strictly before temporal.end.")

        # Warn if the (single) seasonal window is clipped at global boundaries
        if isinstance(self.cadence, SeasonsCadence):
            m1, d1 = map(int, self.cadence.start.split("-"))
            m2, d2 = map(int, self.cadence.end.split("-"))

            for year in range(self.start.year - 1, self.end.year + 1):
                a = _safe_ymd(year, m1, d1)  # nominal season start
                b_same = _safe_ymd(year, m2, d2)  # same-year end, if not wrapping
                win_start = a
                win_end = b_same if a < b_same else _safe_ymd(year + 1, m2, d2)

                # clip to [start, end)
                clipped_start = max(win_start, self.start)
                clipped_end = min(win_end, self.end)

                # if it overlaps and clipping happened -> warn once
                if clipped_start < clipped_end and (
                    clipped_start > win_start or clipped_end < win_end
                ):
                    logger.warning(
                        f"Season ({self.cadence.start} until {self.cadence.end}) near {year} "
                        f"is truncated by [{self.start.date()} — {self.end.date()}]. "
                        "First or last seasonal composite may be partial.",
                    )
                    break
        return self


# ----------------- Export options -----------------
class ExportOpts(BaseModel):
    # only allow these three
    destination: Literal["asset", "drive", "gcs"]

    # destination-specific
    collection_path: str | None = None  # required if asset
    folder: str | None = None  # required if drive; optional prefix if gcs
    bucket: str | None = None  # required if gcs

    # common
    project_id: str | None = None  # GEE project ID for exports
    filename_prefix: str = "biophys"
    crs: str  # e.g. "EPSG:4326" / or specify LOCAL_UTM (automatic)
    scale: PositiveInt | None = None  # meters
    max_pixels: int = Field(default=100_000_000_000, ge=1)

    # Enforce that *no other keys* are accepted
    model_config = ConfigDict(extra="forbid")

    @field_validator("crs")
    def check_epsg(cls, v):
        if not (re.match(r"^EPSG:\d{4,5}$", v) or v == "LOCAL_UTM"):
            raise ValueError("crs must look like 'EPSG:4326' or be 'LOCAL_UTM'.")
        return v

    @model_validator(mode="after")
    def check_destination_requirements(self):
        if self.destination == "asset":
            if not self.collection_path:
                raise ValueError(
                    "When destination='asset', 'collection_path' must be provided.",
                )
        elif self.destination == "drive":
            if not self.folder:
                raise ValueError("When destination='drive', 'folder' must be provided.")
        elif self.destination == "gcs":
            if not self.bucket:
                raise ValueError("When destination='gcs', 'bucket' must be provided.")
            # GCS object prefix is optional; if present, must be str
            if self.folder is not None and not isinstance(self.folder, str):
                raise ValueError(
                    "When destination='gcs', 'folder' must be a string if provided.",
                )
        return self


def _latlon_to_utm_epsg(lat: float, lon: float) -> int:
    """
    Returns the UTM EPSG code for a given latitude and longitude,
    including special cases for Norway and Svalbard.
    EPSG codes:
        - Northern Hemisphere: 32601 to 32660
        - Southern Hemisphere: 32701 to 32760
    """
    long_temp = lon
    zone_number = math.floor((long_temp + 180) / 6) + 1

    # Special case for Norway
    if 56.0 <= lat < 64.0 and 3.0 <= long_temp < 12.0:
        zone_number = 32

    # Special zones for Svalbard
    if 72.0 <= lat < 84.0:
        if 0.0 <= long_temp < 9.0:
            zone_number = 31
        elif 9.0 <= long_temp < 21.0:
            zone_number = 33
        elif 21.0 <= long_temp < 33.0:
            zone_number = 35
        elif 33.0 <= long_temp < 42.0:
            zone_number = 37

    # Determine EPSG code
    if lat >= 0:
        epsg_code = 32600 + zone_number  # Northern hemisphere
    else:
        epsg_code = 32700 + zone_number  # Southern hemisphere

    return epsg_code


# ----------------- Variables -----------------
class Variables(BaseModel):
    model: Literal["s2biophys", "sl2p"] = "s2biophys"
    variable: Literal["laie", "fapar", "fcover"] = "laie"
    bands: list[Literal["mean", "stdDev", "count"]] = ["mean", "stdDev", "count"]

    # Enforce that *no other keys* are accepted
    model_config = ConfigDict(extra="forbid")

    @field_validator("model", "variable", mode="before")
    def lowercase_enums(cls, v):
        return v.lower() if isinstance(v, str) else v


# ----------------- Options -----------------
class Options(BaseModel):
    max_cloud_cover: int = Field(default=50, ge=0, le=100)
    csplus_band: Literal["cs", "cs_cdf"] = "cs"
    cs_plus_threshold: float = Field(default=0.70, ge=0.0, le=1.0)
    clip_min_max: bool = True

    # Enforce that *no other keys* are accepted
    model_config = ConfigDict(extra="forbid")


# ----------------- Top-level -----------------
class ConfigParams(BaseModel):
    spatial: Spatial
    temporal: Temporal
    variables: Variables
    export: ExportOpts
    options: Options
    version: str = Field(default="v02", pattern=r"^v\d{2}$")

    @model_validator(mode="after")
    def initialize_ee(self):
        """Initialize Earth Engine with the specified project ID, if provided."""
        import ee

        if self.export.project_id:
            ee.Initialize(project=self.export.project_id)
        else:
            ee.Initialize()
        return self

    # if export.crs == "LOCAL_UTM", resolve to EPSG code based on spatial geometry and return updated instance
    @model_validator(mode="after")
    def export_crs_resolved(self):
        if self.export.crs != "LOCAL_UTM":
            return self  # Return self if no change is made

        geom = self.spatial.ee_geometry
        centroid = geom.centroid(1).coordinates().getInfo()
        lon, lat = centroid[0], centroid[1]
        epsg_code = _latlon_to_utm_epsg(lat, lon)

        # **This line is necessary to persist the change**
        self.export.crs = f"EPSG:{epsg_code}"

        return self  # Return the updated instance



"""gee_biophys: Tools for biophysical variable modeling and prediction with Google Earth Engine (GEE).

This package provides:
- Data preprocessing and Sentinel-2 input preparation (`s2_input`)
- Model prediction and export tools (`s2_predict`, `s2_export`)
- Configuration utilities (`config`)
- Helper functions for GEE asset management (`utils`, `utils_predict`)
- Command-line interface (`cli`)

Typical usage (as a command-line tool):
    gee-biophys run --config example_configs/minimal_example.yaml

Alternative usage (from within Python, e.g. for debugging):
    from gee_biophys.cli import run
    from pathlib import Path
    run(config=Path("example_configs/minimal_example.yaml"))

Author: Felix Specker
License: MIT
"""
